@@ -10,14 +10,11 @@ from http.client import responses
 from jinja2 import TemplateNotFound
 from tornado import web
 from tornado.httputil import url_concat
-from tornado.httputil import urlparse
 
 from .. import __version__
-from .. import orm
 from ..metrics import SERVER_POLL_DURATION_SECONDS
 from ..metrics import ServerPollStatus
-from ..pagination import Pagination
-from ..utils import admin_only
+from ..scopes import needs_scope
 from ..utils import maybe_future
 from ..utils import url_path_join
 from .base import BaseHandler
@@ -240,7 +237,7 @@ class SpawnHandler(BaseHandler):
             raise web.HTTPError(400, "%s is already running" % (spawner._log_name))
         elif spawner.pending:
             raise web.HTTPError(
-                400, "%s is pending %s" % (spawner._log_name, spawner.pending)
+                400, f"{spawner._log_name} is pending {spawner.pending}"
             )
 
         form_options = {}
@@ -311,10 +308,13 @@ class SpawnHandler(BaseHandler):
         # otherwise it may cause a redirect loop
         if f.done() and f.exception():
             exc = f.exception()
+            self.log.exception(f"Error starting server {spawner._log_name}: {exc}")
+            if isinstance(exc, web.HTTPError):
+                # allow custom HTTPErrors to pass through
+                raise exc
             raise web.HTTPError(
                 500,
-                "Error in Authenticator.pre_spawn_start: %s %s"
-                % (type(exc).__name__, str(exc)),
+                f"Unhandled error starting server {spawner._log_name}",
             )
         return self.redirect(pending_url)
 
@@ -349,9 +349,7 @@ class SpawnPendingHandler(BaseHandler):
                 raise web.HTTPError(404, "No such user: %s" % for_user)
 
         if server_name and server_name not in user.spawners:
-            raise web.HTTPError(
-                404, "%s has no such server %s" % (user.name, server_name)
-            )
+            raise web.HTTPError(404, f"{user.name} has no such server {server_name}")
 
         spawner = user.spawners[server_name]
 
@@ -455,84 +453,22 @@ class AdminHandler(BaseHandler):
     """Render the admin page."""
 
     @web.authenticated
-    @admin_only
+    # stacked decorators: all scopes must be present
+    # note: keep in sync with admin link condition in page.html
+    @needs_scope('admin:users')
+    @needs_scope('admin:servers')
     async def get(self):
-        pagination = Pagination(url=self.request.uri, config=self.config)
-        page, per_page, offset = pagination.get_page_args(self)
-
-        available = {'name', 'admin', 'running', 'last_activity'}
-        default_sort = ['admin', 'name']
-        mapping = {'running': orm.Spawner.server_id}
-        for name in available:
-            if name not in mapping:
-                table = orm.User if name != "last_activity" else orm.Spawner
-                mapping[name] = getattr(table, name)
-
-        default_order = {
-            'name': 'asc',
-            'last_activity': 'desc',
-            'admin': 'desc',
-            'running': 'desc',
-        }
-
-        sorts = self.get_arguments('sort') or default_sort
-        orders = self.get_arguments('order')
-
-        for bad in set(sorts).difference(available):
-            self.log.warning("ignoring invalid sort: %r", bad)
-            sorts.remove(bad)
-        for bad in set(orders).difference({'asc', 'desc'}):
-            self.log.warning("ignoring invalid order: %r", bad)
-            orders.remove(bad)
-
-        # add default sort as secondary
-        for s in default_sort:
-            if s not in sorts:
-                sorts.append(s)
-        if len(orders) < len(sorts):
-            for col in sorts[len(orders) :]:
-                orders.append(default_order[col])
-        else:
-            orders = orders[: len(sorts)]
-
-        # this could be one incomprehensible nested list comprehension
-        # get User columns
-        cols = [mapping[c] for c in sorts]
-        # get User.col.desc() order objects
-        ordered = [getattr(c, o)() for c, o in zip(cols, orders)]
-
-        query = self.db.query(orm.User).outerjoin(orm.Spawner).distinct(orm.User.id)
-        subquery = query.subquery("users")
-        users = (
-            self.db.query(orm.User)
-            .select_entity_from(subquery)
-            .outerjoin(orm.Spawner)
-            .order_by(*ordered)
-            .limit(per_page)
-            .offset(offset)
-        )
-
-        users = [self._user_from_orm(u) for u in users]
-
-        running = []
-        for u in users:
-            running.extend(s for s in u.spawners.values() if s.active)
-
-        pagination.total = query.count()
-
         auth_state = await self.current_user.get_auth_state()
         html = await self.render_template(
             'admin.html',
             current_user=self.current_user,
             auth_state=auth_state,
             admin_access=self.settings.get('admin_access', False),
-            users=users,
-            running=running,
-            sort={s: o for s, o in zip(sorts, orders)},
             allow_named_servers=self.allow_named_servers,
             named_server_limit_per_user=self.named_server_limit_per_user,
-            server_version='{} {}'.format(__version__, self.version_hash),
-            pagination=pagination,
+            server_version=f'{__version__} {self.version_hash}',
+            api_page_limit=self.settings["api_page_default_limit"],
+            base_url=self.settings["base_url"],
         )
         self.finish(html)
 
@@ -550,36 +486,32 @@ class TokenPageHandler(BaseHandler):
             return (token.last_activity or never, token.created or never)
 
         now = datetime.utcnow()
-        api_tokens = []
-        for token in sorted(user.api_tokens, key=sort_key, reverse=True):
-            if token.expires_at and token.expires_at < now:
-                self.db.delete(token)
-                self.db.commit()
-                continue
-            api_tokens.append(token)
 
         # group oauth client tokens by client id
-        # AccessTokens have expires_at as an integer timestamp
-        now_timestamp = now.timestamp()
-        oauth_tokens = defaultdict(list)
-        for token in user.oauth_tokens:
-            if token.expires_at and token.expires_at < now_timestamp:
-                self.log.warning("Deleting expired token")
+        all_tokens = defaultdict(list)
+        for token in sorted(user.api_tokens, key=sort_key, reverse=True):
+            if token.expires_at and token.expires_at < now:
+                self.log.warning(f"Deleting expired token {token}")
                 self.db.delete(token)
                 self.db.commit()
                 continue
             if not token.client_id:
                 # token should have been deleted when client was deleted
-                self.log.warning("Deleting stale oauth token for %s", user.name)
+                self.log.warning("Deleting stale oauth token {token}")
                 self.db.delete(token)
                 self.db.commit()
                 continue
-            oauth_tokens[token.client_id].append(token)
+            all_tokens[token.client_id].append(token)
 
+        # individually list tokens issued by jupyterhub itself
+        api_tokens = all_tokens.pop("jupyterhub", [])
+
+        # group all other tokens issued under their owners
         # get the earliest created and latest last_activity
         # timestamp for a given oauth client
         oauth_clients = []
-        for client_id, tokens in oauth_tokens.items():
+
+        for client_id, tokens in all_tokens.items():
             created = tokens[0].created
             last_activity = tokens[0].last_activity
             for token in tokens[1:]:
@@ -592,8 +524,9 @@ class TokenPageHandler(BaseHandler):
             token = tokens[0]
             oauth_clients.append(
                 {
-                    'client': token.client,
-                    'description': token.client.description or token.client.identifier,
+                    'client': token.oauth_client,
+                    'description': token.oauth_client.description
+                    or token.oauth_client.identifier,
                     'created': created,
                     'last_activity': last_activity,
                     'tokens': tokens,
